import os
import typing
import morpfw
from pathlib import Path
import datetime
from uuid import uuid4
from .base import BlobStorage, Blob
import typing
import json
import morepath

# 1MB
WRITE_BUFF_SIZE = 1073741824


class FSBlob(Blob):

    def __init__(self, path, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.path = path

    def open(self):
        return open(self.path, 'rb')

    def get_size(self):
        stat = os.stat(self.path)
        return stat.st_size


class FSBlobStorage(BlobStorage):

    def __init__(self, request: morepath.Request, path: str):
        self.path = path
        self.request = request

    def _uuid_path(self, uuid):
        first = uuid[:2]
        second = uuid[2:4]
        return os.path.join(self.path, first, second)

    def _meta_path(self, obj_path):
        return '%s.metadata.json' % obj_path

    def put(self, fileobj: typing.BinaryIO,
            filename: str, mimetype: typing.Optional[str] = None,
            size: typing.Optional[int] = None,
            encoding: typing.Optional[str] = None,
            uuid: typing.Optional[str] = None) -> FSBlob:

        if uuid is None:
            uuid = uuid4().hex

        obj_path = os.path.join(self._uuid_path(uuid), uuid)
        meta_path = self._meta_path(obj_path)
        Path(self._uuid_path(uuid)).mkdir(parents=True, exist_ok=True)

        fileobj.seek(0)
        with open(obj_path, 'wb') as o:
            while True:
                data = fileobj.read(WRITE_BUFF_SIZE)
                if not data:
                    break
                o.write(data)

        meta = {
            'uuid': uuid,
            'filename': filename,
            'mimetype': mimetype,
            'size': size,
            'encoding': encoding
        }
        with open(meta_path, 'w') as mo:
            mo.write(json.dumps(meta))

        meta['path'] = obj_path
        return FSBlob(**meta)

    def get(self, uuid: str) -> typing.Optional[Blob]:
        obj_path = os.path.join(self._uuid_path(uuid), uuid)
        meta_path = self._meta_path(obj_path)

        if not os.path.exists(meta_path):
            return None

        meta: dict = {}
        with open(meta_path) as mo:
            meta = json.loads(mo.read())

        meta['path'] = obj_path
        return FSBlob(**meta)

    def delete(self, uuid: str):
        obj_path = os.path.join(self._uuid_path(uuid), uuid)
        meta_path = self._meta_path(obj_path)

        if os.path.exists(obj_path):
            os.unlink(obj_path)

        if os.path.exists(meta_path):
            os.unlink(meta_path)



from morpfw.crud.storage.sqlstorage import Base
from morpfw.crud.storage.sqlstorage import GUID
import sqlalchemy as sa
from sqlalchemy.orm import relationship
import sqlalchemy.orm as saorm
import sqlalchemy_jsonfield as sajson


class User(Base):

    __tablename__ = 'authmanager_users'

    username = sa.Column(sa.String(length=256))
    email = sa.Column(sa.String)
    password = sa.Column(sa.String(length=1024))
    source = sa.Column(sa.String(length=1024))


class APIKey(Base):

    __tablename__ = 'authmanager_apikey'

    userid = sa.Column(GUID)
    label = sa.Column(sa.String(length=256))
    api_identity = sa.Column(sa.String(length=40))
    api_secret = sa.Column(sa.String(length=40))


class Group(Base):

    __tablename__ = 'authmanager_groups'

    groupname = sa.Column(sa.String(length=256))
    memberships = relationship('Membership', cascade='all')


class Membership(Base):

    __tablename__ = 'authmanager_membership'

    group_id = sa.Column(sa.ForeignKey(
        'authmanager_groups.id'))
    user_id = sa.Column(sa.ForeignKey(
        'authmanager_users.id'))
    sa.UniqueConstraint('group_id', 'user_id')

    roles_assignment = relationship('RoleAssignment', cascade='all')


class RoleAssignment(Base):

    __tablename__ = 'authmanager_roleassignment'

    membership_id = sa.Column(sa.ForeignKey('authmanager_membership.id'))
    rolename = sa.Column(sa.String)
    sa.UniqueConstraint('membership_id', 'role_id')



from argh import arg, dispatch_commands
import importlib
import os
import sys
from .main import create_app
from .main import create_admin
import morepath
import yaml
import morpfw
import getpass
import socket
from datetime import datetime


def load(app_path, settings_file, host=None, port=None):
    settings = yaml.load(open(settings_file))

    if not app_path:
        if 'application' not in settings:
            print("'application' section is required in settings")
            sys.exit(1)
        if 'app' not in settings['application']:
            print("Missing application:app entry in settings")
        app_path = settings['application']['app']

    if 'server' in settings:
        if not host:
            host = settings['server'].get('listen_host', '127.0.0.1')
        if not port:
            port = settings['server'].get('listen_port', 5432)

    sys.path.append(os.getcwd())
    mod, clsname = app_path.split(':')
    app_cls = getattr(importlib.import_module(mod), clsname)

    return {
        'app_cls': app_cls,
        'settings': settings,
        'host': host,
        'port': port
    }


@arg('-a', '--app', required=False, default=None, help='Path to App class')
@arg('-s', '--settings', required=True, help='Path to settings.yml')
@arg('-h', '--host', default=None, help='Host')
@arg('-p', '--port', default=None, type=int, help='Port')
def start(app=None, settings=None, host=None, port=None):
    param = load(app, settings, host, port)
    morpfw.run(param['app_cls'], settings=param['settings'], host=param['host'],
               port=param['port'], ignore_cli=True)


@arg('-a', '--app', required=False, help='Path to App class')
@arg('-s', '--settings', required=True, help='Path to settings.yml')
def solo_worker(app=None, settings=None):
    param = load(app, settings)
    hostname = socket.gethostname()
    ws = param['settings']['worker']['celery_worker_settings']
    now = datetime.utcnow().strftime(r'%Y%m%d%H%M')
    worker = param['app_cls'].celery.Worker(
        hostname='worker%s.%s' % (now, hostname), **ws)
    worker.start()


@arg('-a', '--app', required=False, help='Path to App class')
@arg('-s', '--settings', required=True, help='Path to settings.yml')
def scheduler(app=None, settings=None):
    param = load(app, settings)
    hostname = socket.gethostname()
    ss = param['settings']['worker']['celery_scheduler_settings']
    sched = param['app_cls'].celery.Beat(
        hostname='scheduler.%s' % hostname, **ss)
    sched.start()


@arg('-a', '--app', required=False, help='Path to App class')
@arg('-s', '--settings', required=True, help='Path to settings.yml')
@arg('-u', '--username', required=True, help='Username')
@arg('-e', '--email', required=True, help='Email address')
def register_admin(app=None, settings=None, username=None, email=None):
    param = load(app, settings)
    password = getpass.getpass('Enter password for %s: ' % username)
    app = create_app(param['app_cls'], param['settings'])
    while not isinstance(app, morepath.App):
        wrapped = getattr(app, 'app', None)
        if wrapped:
            app = wrapped
        else:
            raise ValueError(
                'Unable to locate app object from middleware')
    create_admin(app=app, username=username,
                 password=password, email=email)


def run():
    dispatch_commands([start, solo_worker, register_admin, scheduler])



from setuptools import setup, find_packages
import sys
import os

version = '0.2.1a2'

long_description = open(
    os.path.join(os.path.dirname(__file__), 'README.rst')).read()

setup(name='morpfw',
      version=version,
      description="Web framework based on morepath",
      long_description=long_description,
      classifiers=[],  # Get strings from http://pypi.python.org/pypi?%3Aaction=list_classifiers
      keywords='',
      author='Izhar Firdaus',
      author_email='izhar@abyres.net',
      url='http://github.com/morpframework/morpfw',
      license='MIT',
      packages=find_packages(exclude=['ez_setup', 'examples', 'tests']),
      include_package_data=True,
      zip_safe=False,
      install_requires=[
          # -*- Extra requirements: -*-
          'argh',
          'rulez>=0.1.1,<0.2.0',
          'more.cors',
          'celery',
          'redis',
          'jsl',
          'pyyaml',
          'more.jsonschema',
          'jsonobject==0.9.4',
          'sqlalchemy',
          'more.signals',
          'DateTime',
          'transitions',
          'jsonpath_ng',
          'python-dateutil',
          'more.jwtauth',
          'more.itsdangerous',
          'sqlsoup',
          'celery',
          'gunicorn',
          'itsdangerous',
          'pyyaml',
          'passlib',
          'jsonschema',
          'more.transaction',
          'zope.sqlalchemy',
          'python-dateutil',
          'more.cors',
          'sqlalchemy_jsonfield',
          'sqlsoup',
          'celery',
          'gunicorn',
          'itsdangerous',
          'pyyaml',
          'passlib',
          'jsonschema',
          'more.transaction',
          'zope.sqlalchemy',
          'more.basicauth',
          'cryptography',
          'elasticsearch>=5.0.0,<6.0.0',
          'pamela',
      ],
      extras_require={
          'test': [
              'nose',
              'webtest',
              'pytest==3.10.1',
              'pytest-html',
              'pytest_postgresql',
              'pytest_rabbitmq',
              'pytest-annotate',
              'pytest-cov',
              'pika',
          ]
      },
      entry_points={
          'morepath': ['scan=morpfw'],
          'console_scripts': [
              'morpfw=morpfw.cli:run'
          ]
      }
      )




class UnprocessableError(Exception):
    def __init__(self, message=None):
        self.message = message or ''


class ValidationError(Exception):
    def __init__(self, field_errors=None, form_errors=None):
        super(ValidationError, self).__init__("Schema validation error")
        self.field_errors = field_errors or []
        self.form_errors = form_errors or []


class FormValidationError(Exception):

    def __init__(self, message):
        self.message = message


class FieldValidationError(Exception):
    def __init__(self, message):
        self.message = message


class AlreadyExistsError(Exception):

    def __init__(self, message):
        self.message = message


class StateUpdateProhibitedError(Exception):
    pass


class BlobStorageNotImplementedError(Exception):
    pass



import os
import morepath
import yaml
from webtest import TestApp as Client
from more.jwtauth import JWTIdentityPolicy
from more.basicauth import BasicAuthIdentityPolicy
from morpfw.main import create_app
from morpfw.main import create_admin
from morpfw.authn.pas.exc import UserExistsError

DEFAULT_SETTINGS = {
    'application': {
        'authn_policy': 'morpfw.authn.noauth:AuthnPolicy',
        'dburi': 'postgresql://postgres@localhost:45678/morp_tests'
    },
    'worker': {
        'enabled': True,
        'celery_name': 'morp_tasks',
        'celery_settings':  {
            'metastore': 'sqlstorage',
            'broker_url': 'amqp://guest:guest@localhost:38567/',
            'result_backend': 'rpc://'
        }
    }
}


def get_settings(config):
    if isinstance(config, str):
        configpath = os.path.join(os.path.dirname(__file__), config)
        if not os.path.exists(configpath):
            return DEFAULT_SETTINGS
        with open(os.path.join(os.path.dirname(__file__), config)) as f:
            settings = yaml.load(f)
    else:
        settings = config
    return settings


def get_client(app, config='settings.yml', **kwargs):
    settings = get_settings(config)
    appobj = create_app(app, settings, **kwargs)
    appobj.initdb()
#	try:
#		create_admin(appobj, username='defaultuser',
#					 password='password', email='admin@localhost.localdomain')
#	except UserExistsError:
#		pass
    c = Client(appobj)
    return c



from morpfw.crud.util import jsonobject_to_jsl
from morpfw.crud.util import jsl_to_jsonobject
import jsonobject
import jsl


class SubSchema(jsonobject.JsonObject):
    stringfield = jsonobject.StringProperty()
    intfield = jsonobject.IntegerProperty()


class SampleSchema(jsonobject.JsonObject):
    stringfield = jsonobject.StringProperty()
    intfield = jsonobject.IntegerProperty()
    dictfield = jsonobject.DictProperty()
    arrayfield = jsonobject.ListProperty()
    documentarrayfield = jsonobject.ListProperty(item_type=SubSchema)
    documentfield = jsonobject.DictProperty(item_type=SubSchema)
    datefield = jsonobject.DateProperty()
    datetimefield = jsonobject.DateTimeProperty()


def test_jsonobject_convert():
    schema = jsonobject_to_jsl(SampleSchema)
    field = getattr(schema, 'stringfield')
    assert isinstance(field, jsl.StringField)
    assert field.name == 'stringfield'

    field = getattr(schema, 'intfield')
    assert isinstance(field, jsl.IntField)
    assert field.name == 'intfield'

    field = getattr(schema, 'dictfield')
    assert isinstance(field, jsl.DictField)
    assert field.name == 'dictfield'

    field = getattr(schema, 'arrayfield')
    assert isinstance(field, jsl.ArrayField)
    assert field.name == 'arrayfield'

    field = getattr(schema, 'documentarrayfield')
    assert isinstance(field, jsl.ArrayField)
    assert isinstance(field.items, jsl.DocumentField)
    assert isinstance(field.items.document_cls.stringfield, jsl.StringField)

    field = getattr(schema, 'documentfield')
    assert isinstance(field, jsl.DocumentField)
    assert isinstance(field.document_cls.stringfield, jsl.StringField)

    field = getattr(schema, 'datetimefield')
    assert isinstance(field, jsl.DateTimeField)

    field = getattr(schema, 'datefield')
    assert isinstance(field, jsl.DateTimeField)

    assert schema.get_schema()

    schema = jsl_to_jsonobject(schema)



from .model import APIKeyCollection, APIKeyModel, APIKeySchema
from ..app import App


def get_apikey_collection(request):
    authn_provider = request.app.get_authn_provider(request)
    storage = authn_provider.get_storage(APIKeyModel, request)
    return APIKeyCollection(request, storage)


def get_apikey(request, identifier):
    collection = get_apikey_collection(request)
    return collection.get(identifier)


@App.path(model=APIKeyModel,
          path='apikey/{uuid}')
def _get_apikey(app, request, uuid):
    return get_apikey(request, uuid)


@App.path(model=APIKeyCollection,
          path='apikey')
def _get_apikey_collection(app, request):
    return get_apikey_collection(request)



import nose
import morepath
from webtest import TestApp as Client
from morpfw.authn.pas.app import App
from morpfw.main import create_app
from morpfw.main import create_admin
from morpfw.authn.pas.user.model import UserCollection, UserSchema, GroupSchema
from more.jwtauth import JWTIdentityPolicy
import json
import yaml
import os
import copy
from more.basicauth import BasicAuthIdentityPolicy
import time
import transaction


def get_client(app, config='settings.yml', **kwargs):
    if isinstance(config, str):
        with open(os.path.join(os.path.dirname(__file__), config)) as f:
            settings = yaml.load(f)
    else:
        settings = config

    kwargs = {}
    appobj = create_app(app, settings, **kwargs)
    request = appobj.request_class(
        app=appobj, environ={'PATH_INFO': '/'})

    username = 'admin'
    password = 'password'
    create_admin(appobj, username, password,
                 '%s@localhost.localdomain' % username)
    transaction.commit()
    c = Client(appobj)
    return c


def login(c, username, password='password'):

    r = c.post_json('/auth/user/+login', {
        'username': username,
        'password': password
    })

    assert r.status_code == 200

    token = r.headers.get('Authorization').split()[1]

    c.authorization = ('JWT', token)

    return r.json


def logout(c):
    c.authorization = None


def _test_authentication(c):
    r = c.get('/auth/user/+login')

    # test schema access

#    assert r.json['schema']['title'] == 'credential'

    ll = r.json['links'][0]
    assert ll['rel'] == 'login'
    assert ll['type'] == 'POST'

    r = c.post_json(ll['href'], {
        'username': 'admin',
        'password': 'password'
    })

    assert r.json == {
        'status': 'success'
    }

    logout(c)

    # test wrong login
    r = c.post_json('/auth/user/+login', {
        'username': 'invaliduser',
        'password': 'invalidpassword'
    }, expect_errors=True)

    assert r.json == {
        'status': 'error',
        'error': {
            'code': 401,
            'message': 'Invalid Username / Password'
        }
    }

    r = c.post_json('/auth/user/+login', {
        'username': 'admin',
        'password': 'invalidpassword'
    }, expect_errors=True)

    assert r.json == {
        'status': 'error',
        'error': {
            'code': 401,
            'message': 'Invalid Username / Password'
        }
    }

    # from now on we login as admin

    login(c, 'admin')

    # test refreshing token
    time.sleep(2)

    r = c.get('/auth/user/+refresh_token')

    n = r.headers.get('Authorization').split()

    assert c.authorization[1] != n[1]

    c.authorization = ('JWT', n[1])

    r = c.get('/auth/user/admin')

    assert r.json['data']['username'] == 'admin'
    assert list(
        [l['name'] for l in r.json['links'] if l['rel'] == 'group']
    ) == ['__default__']
    assert r.json['data']['state'] == 'active'

    # query for nonexistent user

    r = c.get('/auth/user/unknownuser', expect_errors=True)

    assert r.status_code == 404

    r = c.get('/auth/user/')

#    assert r.json['schema']['title'] == 'user'

    logout(c)

    # register new user, you shouldnt be allowed to if not logged in
    # as admin

    r = c.post_json('/auth/user/+register',
                    {'username': 'user1',
                     'password': 'password',
                     'password_validate': 'password'}, expect_errors=True)

    assert r.status_code == 403

    # register new user

    login(c, 'admin')

    r = c.post_json('/auth/user/+register',
                    {'username': 'user1',
                     'email': 'user1@localhost.com',
                     'password': 'password',
                     'password_validate': 'password'})

    assert r.json == {'status': 'success'}

    r = c.post_json('/auth/user/+register',
                    {'username': 'user2',
                     'email': 'user2@localhost.com',
                     'password': 'password',
                     'password_validate': 'password'})

    assert r.json == {'status': 'success'}

    # fail if password is not string
    r = c.post_json('/auth/user/+register',
                    {'username': 'user3',
                     'email': 'user3@localhost.com',
                     'password': {'hello': 'world'}},
                    expect_errors=True)

    assert r.status_code == 422

    # fail if duplicate user

    r = c.post_json('/auth/user/+register',
                    {'username': 'user1',
                     'email': 'user1@localhost.com',
                     'password': 'password',
                     'password_validate': 'password'},
                    expect_errors=True)

    assert r.status_code == 422
    assert r.json['status'] == 'error'
    assert r.json['type'] == 'UserExistsError'

    # fail if invalid email

    r = c.post_json('/auth/user/+register',
                    {'username': 'user4',
                     'email': 'user4',
                     'password': 'password',
                     'password_validate': 'password'},
                    expect_errors=True)

    assert r.status_code == 422

    r = c.post_json('/auth/user/+register',
                    {'username': 'user4',
                     'email': 'user4@localhost',
                     'password': 'password',
                     'password_validate': 'password'},
                    expect_errors=True)

    assert r.status_code == 422

    r = c.get('/auth/user/user1')

    assert r.json['data']['username'] == 'user1'
    assert list(
        [l['name'] for l in r.json['links'] if l['rel'] == 'group']
    ) == ['__default__']

    assert r.json['data']['state'] == 'active'
    assert 'password' not in r.json['data'].keys()

    # attempt to login as user1
    login(c, 'user1')

    # attempt to login as user1 with user1 email
    login(c, 'user1@localhost.com')

    login(c, 'admin')
    r = c.post_json(
        '/auth/user/user1/+statemachine', {'transition': 'deactivate'})

    r = c.get('/auth/user/user1')

    assert r.json['data']['username'] == 'user1'
    assert list(
        [l['name'] for l in r.json['links'] if l['rel'] == 'group']
    ) == ['__default__']
    assert r.json['data']['state'] == 'inactive'

    r = c.post_json('/auth/user/+login', {
        'username': 'user1',
        'password': 'password'
    }, expect_errors=True)

    assert r.status_code == 401

    r = c.post_json(
        '/auth/user/user1/+statemachine', {'transition': 'activate'})

    r = c.get('/auth/user/user1')

    assert r.json['data']['username'] == 'user1'
    assert list(
        [l['name'] for l in r.json['links'] if l['rel'] == 'group']
    ) == ['__default__']

    assert r.json['data']['state'] == 'active'

    login(c, 'user1')

    # reject setting password through the update API

    r = c.post_json('/auth/user/user1/',
                    {'password': 'newpass'}, expect_errors=True)

    assert r.status_code == 405

    r = c.patch_json('/auth/user/user1/',
                     {'password': 'newpass'}, expect_errors=True)

    assert r.status_code == 422

    r = c.patch_json('/auth/user/user1/',
                     {'username': 'newusername'}, expect_errors=True)

    assert r.status_code == 422

    login(c, 'admin')

    # admin should be allowed to set password without requiring current password
    r = c.post_json('/auth/user/user1/+change_password', {
        'new_password': 'newpass',
        'new_password_validate': 'newpass'
    })

    assert r.status_code == 200

    login(c, 'user1', 'newpass')

    # user require current password
    r = c.post_json('/auth/user/user1/+change_password', {
        'new_password': 'password',
        'new_password_validate': 'password'
    }, expect_errors=True)

    assert r.status_code == 422

    # user require current password
    r = c.post_json('/auth/user/user1/+change_password', {
        'password': 'newpass',
        'new_password': 'password',
        'new_password_validate': 'password'
    }, expect_errors=True)

    assert r.status_code == 200

    login(c, 'admin')

    r = c.get('/auth/user/admin')

    admin_user = r.json

    # api keys
    r = c.post_json('/auth/apikey/',
                    {'label': 'samplekey', 'password': 'password'})

    key_identity = r.json['data']['api_identity']
    key_secret = r.json['data']['api_secret']
    key_uuid = r.json['data']['uuid']
    assert len(key_identity) == 32
    assert len(key_secret) == 32
    assert r.json['data']['userid'] == admin_user['data']['uuid']

    r = c.get('/auth/apikey/%s' % key_uuid)

    assert key_identity == r.json['data']['api_identity']
    assert key_secret == r.json['data']['api_secret']
    assert key_uuid == r.json['data']['uuid']

    # user1 shouldnt see admin's apikey

    login(c, 'user1')

    r = c.get('/auth/apikey/+search')

    assert len(r.json['results']) == 0

    logout(c)

    # lets try deactivating user1 using API key
    r = c.post_json(
        '/auth/user/user1/+statemachine', {'transition': 'deactivate'},
        expect_errors=True)

    assert r.status_code == 403

    r = c.post_json(
        '/auth/user/user1/+statemachine',
        {'transition': 'deactivate'}, headers=[
            ('X-API-KEY', '.'.join([key_identity, key_secret]))
        ])

    assert r.status_code == 200

    # activate back
    r = c.post_json(
        '/auth/user/user1/+statemachine',
        {'transition': 'activate'}, headers=[
            ('X-API-KEY', '.'.join([key_identity, key_secret]))
        ])

    assert r.status_code == 200

    login(c, 'admin')

    r = c.get('/auth/group/')

#    assert r.json['schema']['title'] == 'group'

    r = c.post_json('/auth/group/', {'groupname': 'group1'})

    assert r.json['data']['groupname'] == 'group1'

    r = c.post_json('/auth/group/', {'groupname': 'group1'},
                    expect_errors=True)

    assert r.json['status'] == 'error'
    assert r.json['type'] == 'GroupExistsError'

    r = c.get('/auth/group/group1')

    assert r.json['data']['groupname'] == 'group1'

    r = c.post_json('/auth/group/group1/+grant',
                    {'mapping': [{
                        'user': {
                            'username': 'user1'
                        },
                        'roles': ['member']}]})

    assert r.json == {'status': 'success'}

    r = c.post_json('/auth/group/group1/+grant',
                    {'mapping': [{
                        'user': {
                            'username': 'dummyuser'
                        }, 'roles': ['member']
                    }]},
                    expect_errors=True)

    assert r.status_code == 422

    r = c.get('/auth/user/user1')

    assert list(sorted(
        [l['name'] for l in r.json['links'] if l['rel'] == 'group']
    )) == ['__default__', 'group1']

    r = c.get('/auth/group/group1/+members')

    assert r.json['users'][0]['username'] == 'user1'
    assert r.json['users'][0]['roles'] == ['member']

    r = c.post_json('/auth/group/group1/+grant', {'mapping': [{
        'user': {'username': 'user1'},
        'roles': ['manager']
    }]})

    assert r.json == {'status': 'success'}

    r = c.get('/auth/group/group1/+members')

    assert r.json['users'][0]['roles'] == ['member', 'manager']

    r = c.post_json('/auth/group/group1/+grant', {'mapping': [{
        'user': {'username': 'user1'},
        'roles': ['editor']
    }]})

    assert r.json == {'status': 'success'}

    r = c.get('/auth/group/group1/+members')

    assert sorted(r.json['users'][0]['roles']
                  ) == sorted(['member', 'editor', 'manager'])

    r = c.get('/auth/user/user1/+roles')

    assert sorted(r.json['group1']) == sorted(['member', 'editor', 'manager'])

    r = c.post_json('/auth/group/group1/+revoke',
                    {'mapping': [{
                        'user': {
                            'username': 'user1'
                        },
                        'roles': ['manager']
                    }]})

    assert r.json == {'status': 'success'}

    r = c.get('/auth/group/group1/+members')

    assert r.json['users'][0]['roles'] == ['member', 'editor']

    r = c.post_json('/auth/group/group1/+revoke', {
        'mapping': [{
            'user': {'username': 'user1'},
            'roles': ['editor', 'member']}]
    })

    assert r.json == {'status': 'success'}

    r = c.get('/auth/group/group1/+members')

    assert len(r.json['users']) == 0

    r = c.delete('/auth/user/user1')

    assert r.json == {'status': 'success'}

    r = c.get('/auth/user/user1', expect_errors=True)

    assert r.status_code == 404

    logout(c)

    r = c.delete('/auth/user/user1', expect_errors=True)



from .test_auth import _test_authentication, get_client
from more.jwtauth import JWTIdentityPolicy
from morpfw.app import BaseApp
from morpfw.authn.pas.app import App
import os
import yaml
from morpfw.authn.pas.policy import MemoryStorageAuthnPolicy
from morpfw.authn.pas.policy import MemoryStorageAuthApp as BaseAuthApp
from morpfw.authz.pas import DefaultAuthzPolicy


class MemoryStorageAuthApp(BaseAuthApp, DefaultAuthzPolicy):
    pass


class MemoryStorageApp(BaseApp, DefaultAuthzPolicy):
    pass


class AuthnPolicy(MemoryStorageAuthnPolicy):
    app_cls = MemoryStorageAuthApp


@MemoryStorageApp.mount(app=AuthnPolicy.app_cls, path='/auth')
def mount_app(app):
    return AuthnPolicy.app_cls()


def test_authentication_memorystorage():
    with open(os.path.join(os.path.dirname(__file__),
                           'settings-memorystorage.yml')) as f:
        settings = yaml.load(f)

    c = get_client(MemoryStorageApp, settings)
    _test_authentication(c)



from morpfw import Schema
from ..app import App
from dataclasses import dataclass, field
import typing


@dataclass
class GroupSchema(Schema):
    groupname: typing.Optional[str] = field(default=None, metadata={
        'morpfw': {'required': True}})


@App.identifierfields(schema=GroupSchema)
def group_identifierfields(schema):
    return ['groupname']


@dataclass
class MemberSchema(Schema):
    users: typing.List[str] = field(default_factory=list, metadata={
                                    'morpfw': {'required': True}})



from uuid import uuid4
from ..interfaces import ISchema
from .app import App
from dataclasses import dataclass, field
from datetime import datetime
import typing


@dataclass
class Schema(ISchema):

    id: typing.Optional[int] = None
    uuid: typing.Optional[str] = field(default_factory=lambda: uuid4().hex)
    creator: typing.Optional[str] = None
    created: typing.Optional[datetime] = None
    modified: typing.Optional[datetime] = None
    state: typing.Optional[str] = None
    deleted: typing.Optional[datetime] = None
    blobs: typing.Optional[dict] = field(
        default_factory=dict, metadata={'morpfw': {
            'exclude_if_empty': True
        }})
    xattrs: typing.Optional[dict] = field(
        default_factory=dict, metadata={'morpfw': {
            'exclude_if_empty': True
        }})


@App.identifierfields(schema=Schema)
def identifierfields(schema):
    return ['uuid']


@App.default_identifier(schema=Schema)
def default_identifier(schema, obj, request):
    fields = request.app.get_identifierfields(schema)
    res = []
    for f in fields:
        if f == 'uuid':
            res.append(uuid4().hex)
        else:
            res.append(obj[f])
    return res



from .app import App
from .model import Model


class Adapter(object):

    def __init__(self, context: Model):
        self.context = context
        self.request = context.request
        self.app = context.app

    @property
    def schema(self):
        return self.context.schema

    @property
    def identifier(self):
        return self.context.identifier

    @property
    def data(self):
        return self.context.data

    def json(self):
        return self.context.json()

    def update(self, *args, **kwargs):
        return self.context.update(*args, **kwargs)

    def delete(self):
        return self.context.delete()

    def transform_json(self, data):
        return data

    def links(self):
        return []

    def __repr__(self):
        return "<Adapted %s:(%s)>" % (self.__class__.__name__, self.context)


@App.rulesadapter(model=Model)
def get_default_rulesadapter(obj):
    return Adapter(obj)



class All(object):
    pass


class View(All):
    pass


class Edit(All):
    pass


class Create(All):
    pass


class Delete(All):
    pass


class Search(All):
    pass


class Aggregate(All):
    pass



import jsl
from .crud_common import App as BaseApp
from morpfw.crud.model import Collection, Model
from morpfw.crud.storage.sqlstorage import SQLStorage, Base, GUID
from morpfw.crud.blobstorage.fsblobstorage import FSBlobStorage
from .crud_common import get_client, run_jslcrud_test, PageCollection, PageModel
from .crud_common import ObjectCollection, ObjectModel
from .crud_common import NamedObjectCollection, NamedObjectModel
from .crud_common import BlobObjectCollection, BlobObjectModel
from .crud_common import FSBLOB_DIR
import pprint
from more.transaction import TransactionApp
from morepath.reify import reify
from morepath.request import Request

from sqlalchemy.orm import sessionmaker
from zope.sqlalchemy import register as register_session
import morpfw.crud.signals as signals
import sqlalchemy as sa
import sqlalchemy_jsonfield as sajson
from morpfw.app import SQLApp


class App(BaseApp, SQLApp):
    pass


class Page(Base):

    __tablename__ = 'jslcrud_test_page'

    id = sa.Column(sa.Integer, primary_key=True, autoincrement=True)
    title = sa.Column(sa.String(length=1024), default='')
    body = sa.Column(sa.Text(), default='')
    value = sa.Column(sa.Integer)
    footer = sa.Column(sa.String(length=1024), default='')


class PageStorage(SQLStorage):
    model = PageModel
    orm_model = Page


@App.path(model=PageCollection, path='pages')
def collection_factory(request):
    storage = PageStorage(request)
    return PageCollection(request, storage)


@App.path(model=PageModel, path='pages/{identifier}')
def model_factory(request, identifier):
    storage = PageStorage(request)
    return storage.get(identifier)


class Object(Base):

    __tablename__ = 'jslcrud_test_object'

    body = sa.Column(sa.String(length=1024), default='')
    created_flag = sa.Column(sa.Boolean, default=False)
    updated_flag = sa.Column(sa.Boolean, default=False)
    attrs = sa.Column(sajson.JSONField)


class ObjectStorage(SQLStorage):
    model = ObjectModel
    orm_model = Object


@App.path(model=ObjectCollection, path='objects')
def object_collection_factory(request):
    storage = ObjectStorage(request)
    return ObjectCollection(request, storage)


@App.path(model=ObjectModel, path='objects/{identifier}')
def object_model_factory(request, identifier):
    storage = ObjectStorage(request)
    return storage.get(identifier)


class NamedObject(Base):

    __tablename__ = 'jslcrud_test_namedobject'

    name = sa.Column(sa.String(length=1024), primary_key=True)
    body = sa.Column(sa.String(length=1024), default='')
    created_flag = sa.Column(sa.Boolean, default=False)
    updated_flag = sa.Column(sa.Boolean, default=False)


class NamedObjectStorage(SQLStorage):
    model = NamedObjectModel
    orm_model = NamedObject


@App.path(model=NamedObjectCollection, path='named_objects')
def namedobject_collection_factory(request):
    storage = NamedObjectStorage(request)
    return NamedObjectCollection(request, storage)


@App.path(model=NamedObjectModel, path='named_objects/{identifier}')
def namedobject_model_factory(request, identifier):
    storage = NamedObjectStorage(request)
    return storage.get(identifier)


class BlobObject(Base):

    __tablename__ = 'jslcrud_test_blobobject'

    blobs = sa.Column(sajson.JSONField())


class BlobObjectStorage(SQLStorage):
    model = BlobObjectModel
    orm_model = BlobObject


@App.storage(model=BlobObjectModel)
def get_blobobject_storage(model, request, blobstorage):
    return BlobObjectStorage(request, blobstorage=blobstorage)


@App.blobstorage(model=BlobObjectModel)
def get_blobobject_blobstorage(model, request):
    return FSBlobStorage(request, FSBLOB_DIR)


@App.path(model=BlobObjectCollection, path='blob_objects')
def blobobject_collection_factory(request):
    storage = request.app.get_storage(BlobObjectModel, request)
    return BlobObjectCollection(request, storage)


@App.path(model=BlobObjectModel, path='blob_objects/{identifier}')
def blobobject_model_factory(request, identifier):
    storage = request.app.get_storage(BlobObjectModel, request)
    return storage.get(identifier)


def test_sqlstorage(pgsql_db):
    c = get_client(App, config='settings-sqlalchemy.yml')
    run_jslcrud_test(c)
